"""
helax (`helax`)

Helax is a package for computing helicity amplitudes. We offer two different
flavors: a pure `NumPy` version and a `jax` version. The `jax` version requires
`jax` and `jaxlib` to be installed. The `jax` version is useful for utilizing a
GPU.
"""
